# Generated by Django 5.0.6 on 2024-08-28 09:41

import django.db.models.deletion
from django.db import migrations, models


def copy_through(apps, schema_editor):
    Portfolio = apps.get_model("wbportfolio", "Portfolio")
    PortfolioBankAccountThroughModel = apps.get_model("wbportfolio", "PortfolioBankAccountThroughModel")

    # We loop over portfolios and their bank accounts with distinct, because we add unique constraints later
    for portfolio in Portfolio.objects.filter(bank_accounts__isnull=False).distinct():
        for bank_account in portfolio.bank_accounts.all().distinct():
            PortfolioBankAccountThroughModel.objects.create(portfolio=portfolio, bank_account=bank_account)


class Migration(migrations.Migration):
    dependencies = [
        ("directory", "0009_remove_entry_external_identfier_and_more"),
        ("wbportfolio", "0065_alter_portfolio_managers_claim_as_shares_and_more"),
    ]

    operations = [
        migrations.CreateModel(
            name="PortfolioBankAccountThroughModel",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                (
                    "portfolio_bank_account_type",
                    models.CharField(
                        choices=[("CASH", "Cash"), ("SECURITIES", "Securities")], default="CASH", max_length=255
                    ),
                ),
                (
                    "bank_account",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="portfolio_through",
                        to="directory.bankingcontact",
                    ),
                ),
                (
                    "portfolio",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="bank_account_through",
                        to="wbportfolio.portfolio",
                    ),
                ),
            ],
            options={
                "verbose_name": "Portfolio Bank Account",
                "verbose_name_plural": "Portfolio Bank Accounts",
            },
        ),
        migrations.RunPython(copy_through),
        migrations.RemoveField(
            model_name="portfolio",
            name="bank_accounts",
        ),
        migrations.AddField(
            model_name="portfolio",
            name="bank_accounts",
            field=models.ManyToManyField(
                blank=True,
                related_name="wbportfolio_portfolios",
                through="wbportfolio.PortfolioBankAccountThroughModel",
                to="directory.bankingcontact",
            ),
        ),
        migrations.AddConstraint(
            model_name="portfoliobankaccountthroughmodel",
            constraint=models.UniqueConstraint(
                fields=("portfolio", "bank_account"), name="unique_portfolio_bank_account"
            ),
        ),
        migrations.AddConstraint(
            model_name="portfoliobankaccountthroughmodel",
            constraint=models.UniqueConstraint(
                condition=models.Q(("portfolio_bank_account_type", "SECURITIES")),
                fields=("portfolio",),
                name="unique_portfolio_bank_account_type",
            ),
        ),
        migrations.AddField(
            model_name="portfolio",
            name="only_weighting",
            field=models.BooleanField(
                default=False,
                help_text="Indicates that this portfolio is only utilizing weights and disregards shares, e.g. a model portfolio",
            ),
        ),
        migrations.AlterField(
            model_name="portfolioportfoliothroughmodel",
            name="type",
            field=models.CharField(
                choices=[
                    ("PRIMARY", "Primary"),
                    ("MODEL", "Model"),
                    ("BENCHMARK", "Benchmark"),
                    ("INDEX", "Index"),
                    ("CUSTODIAN", "Custodian"),
                ],
                default="PRIMARY",
                verbose_name="Type",
            ),
        ),
    ]



from typing import Generator

from wbcompliance.models.risk_management import backend
from wbcompliance.models.risk_management.dispatch import register
from wbcore import serializers as wb_serializers
from wbfdm.models import Instrument, InstrumentList
from wbfdm.serializers.instruments.instrument_lists import (
    InstrumentListRepresentationSerializer,
)

from wbportfolio.pms.typing import Portfolio as PortfolioDTO

from .mixins import ActivePortfolioRelationshipMixin


@register("Instrument List Portfolio Rule Backend", rule_group_key="portfolio")
class RuleBackend(ActivePortfolioRelationshipMixin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.instrument_list_type:
            self.instrument_lists = InstrumentList.objects.filter(instrument_list_type=self.instrument_list_type)

        self.instrument_lists_repr = " ,".join(map(lambda x: x.name, self.instrument_lists))
        self.severity = self.thresholds[0].severity

    @classmethod
    def get_serializer_class(cls) -> wb_serializers.Serializer:
        class RuleBackendSerializer(wb_serializers.Serializer):
            exclude = wb_serializers.BooleanField(
                default=True,
                label="Exclude",
                help_text="If true, the rule will check that the portfolio composition DOES NOT intersect the given instrument list",
            )
            instrument_list_type = wb_serializers.ChoiceField(
                choices=InstrumentList.InstrumentListType.choices,
                required=False,
                default=None,
                allow_null=True,
                help_text="If specified, will dynamically load the list of instrument list to check of the same specified type",
                label="Instrument List Type",
            )
            instrument_lists = wb_serializers.PrimaryKeyRelatedField(
                queryset=InstrumentList.objects.all(),
                many=True,
                default=None,
                allow_null=True,
                label="Instrument Lists",
            )
            _instrument_lists = InstrumentListRepresentationSerializer(many=True)

            @classmethod
            def get_parameter_fields(cls):
                return [
                    "exclude",
                    "instrument_list_type",
                    "instrument_lists",
                ]

        return RuleBackendSerializer

    def _get_breached_instruments(self, instruments: list[Instrument]) -> list[InstrumentList]:
        breached_instruments = []
        for instrument_list in self.instrument_lists:
            if (
                instrument_list.get_instruments(self.evaluation_date, only_validated=True)
                .filter(id__in=instruments.values("id"))
                .exists()
            ):
                breached_instruments.append(instrument_list)
        return breached_instruments

    def _process_dto(self, portfolio: PortfolioDTO, **kwargs) -> Generator[backend.IncidentResult, None, None]:
        for instrument_id in portfolio.positions_map.keys():
            instrument = Instrument.objects.get(id=instrument_id)
            ancestors = instrument.get_ancestors(include_self=True)
            breached_instruments = self._get_breached_instruments(ancestors)

            if self.exclude and len(breached_instruments) > 0:
                report_details = {
                    "Instrument Lists": ", ".join(map(lambda o: o.name, breached_instruments)),
                }
                yield backend.IncidentResult(
                    breached_object=instrument,
                    breached_object_repr=str(instrument),
                    breached_value=f"# {len(breached_instruments)}",
                    report_details=report_details,
                    severity=self.severity,
                )
            elif not self.exclude and len(breached_instruments) == 0:
                report_details = {"Instrument Lists": self.instrument_lists_repr}
                yield backend.IncidentResult(
                    breached_object=instrument,
                    breached_object_repr=str(instrument),
                    breached_value="# 0",
                    report_details=report_details,
                    severity=self.severity,
                )



import pytest
from pandas.tseries.offsets import BDay

from wbportfolio.models import Order


@pytest.mark.django_db
class TestHandleTradeImport:
    def test_skips_cash_trades(self, trade_factory, order_factory, instrument_factory):
        """Ignores cash trades entirely."""
        cash_instrument = instrument_factory.create(is_cash=True)
        order = order_factory.create(underlying_instrument=cash_instrument)
        cash_trade = trade_factory.create(
            underlying_instrument=cash_instrument,
            transaction_date=(order.value_date + BDay(1)).date(),
            portfolio=order.portfolio,
        )  # this is theorically invalid, but it is just for testing purpose

        Order.handle_trade_import(cash_trade)
        order.refresh_from_db()

        assert order.execution_trade is None
        # No changes expected

    def test_skips_customer_trades(self, customer_trade_factory, order_factory):
        """Ignores customer trades."""

        order = order_factory.create()
        customer_trade = customer_trade_factory.create(
            underlying_instrument=order.underlying_instrument,
            transaction_date=(order.value_date + BDay(1)).date(),
            portfolio=order.portfolio,
        )  # this is theorically invalid, but it is just for testing purpose

        Order.handle_trade_import(customer_trade)
        # No changes expected
        assert order.execution_trade is None

    def test_links_to_exact_instrument_match(self, trade_factory, order_factory):
        """Links to earliest order with exact instrument match."""
        order = order_factory.create()
        trade = trade_factory.create(
            underlying_instrument=order.underlying_instrument,
            transaction_date=(order.value_date + BDay(4)).date(),
            portfolio=order.portfolio,
        )

        Order.handle_trade_import(trade, bday_window=2)
        order.refresh_from_db()
        assert order.execution_trade is None

        Order.handle_trade_import(trade, bday_window=3)
        order.refresh_from_db()
        assert order.execution_trade == trade
        assert order.execution_status == Order.ExecutionStatus.EXECUTED

    def test_links_to_parent_instrument_match(self, trade_factory, order_factory, instrument_factory):
        """Links via parent instrument hierarchy."""
        parent = instrument_factory.create()
        child1 = instrument_factory.create(parent=parent)
        child2 = instrument_factory.create(parent=parent)

        order = order_factory.create(
            underlying_instrument=child2,  # Parent match
        )
        trade = trade_factory.create(
            transaction_date=(order.value_date + BDay(1)).date(),
            portfolio=order.portfolio,
            underlying_instrument=child1,
        )

        Order.handle_trade_import(trade)
        order.refresh_from_db()
        assert order.execution_trade == trade

    def test_update_earlier_execution(self, trade_factory, order_factory):
        """Doesn't overwrite earlier execution_trade."""

        order = order_factory.create()

        trade = trade_factory.create(
            transaction_date=(order.value_date + BDay(2)).date(),
            portfolio=order.portfolio,
            underlying_instrument=order.underlying_instrument,
        )
        Order.handle_trade_import(trade)
        order.refresh_from_db()
        assert order.execution_trade == trade  # Unchanged

        earlier_trade = trade_factory.create(
            transaction_date=(order.value_date + BDay(1)).date(),
            portfolio=order.portfolio,
            underlying_instrument=order.underlying_instrument,
        )
        Order.handle_trade_import(earlier_trade)
        order.refresh_from_db()
        assert order.execution_trade == earlier_trade  # Unchanged



# Generated by Django 5.0.12 on 2025-11-11 10:16

import django_fsm
import django.db.models.deletion
from decimal import Decimal
from django.db import migrations, models
import django_fsm
from django.db import migrations

def migrate_status(apps, schema_editor):
    OrderProposal = apps.get_model("wbportfolio", "OrderProposal")
    OrderProposal.objects.filter(status="APPLIED").update(status="CONFIRMED")


class Migration(migrations.Migration):

    dependencies = [
        ('wbportfolio', '0091_remove_order_execution_confirmed_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='order',
            name='quantization_error',
            field=models.DecimalField(decimal_places=8, default=Decimal('0'), max_digits=9, verbose_name='Quantization Error'),
        ),
        migrations.AlterField(
            model_name='orderproposal',
            name='status',
            field=django_fsm.FSMField(choices=[('DRAFT', 'Draft'), ('PENDING', 'Pending'), ('APPROVED', 'Approved'), ('DENIED', 'Denied'), ('EXECUTION', 'Execution'), ('APPLIED', 'Applied'), ('FAILED', 'Failed')], default='DRAFT', max_length=50, verbose_name='Status'),
        ),
        migrations.AlterField(
            model_name='orderproposal',
            name='status',
            field=django_fsm.FSMField(
                choices=[('DRAFT', 'Draft'), ('PENDING', 'Pending'), ('APPROVED', 'Approved'), ('DENIED', 'Denied'),
                         ('EXECUTION', 'Execution'), ('CONFIRMED', 'Confirmed'), ('FAILED', 'Failed')], default='DRAFT',
                max_length=50, verbose_name='Status'),
        ),
        migrations.AddField(
            model_name='order',
            name='execution_trade',
            field=models.OneToOneField(blank=True, help_text='The executed Trade', null=True,
                                       on_delete=django.db.models.deletion.SET_NULL, related_name='order',
                                       to='wbportfolio.trade'),
        ),
        migrations.RunPython(migrate_status),

    ]



from wbcore.contrib.color.enums import WBColor
from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display.view_config import DisplayViewConfig


class DailyPortfolioCashFlowDisplayConfig(DisplayViewConfig):
    def get_instance_display(self) -> dp.Display:
        return dp.create_simple_display(
            grid_template_areas=[
                ["portfolio", "value_date", "cash_flow_forecast", "total_assets"],
                ["estimated_total_assets", "cash", "cash_flow_asset_ratio", "true_cash"],
                ["cash_pct", "true_cash_pct", "target_cash", "excess_cash"],
                ["proposed_rebalancing", "proposed_rebalancing", "rebalancing", "rebalancing"],
                [dp.repeat_field(4, "comment")],
            ]
        )

    def get_list_display(self) -> dp.ListDisplay:
        negative_formatting = dp.FormattingRule(condition=("<", 0), style={"color": WBColor.RED_DARK.value})
        positive_formatting = dp.FormattingRule(condition=(">", 0), style={"color": WBColor.GREEN_DARK.value})
        bold_formatting = dp.FormattingRule(condition=("!=", 0), style={"fontWeight": "bold"})
        pending_formatting = dp.FormattingRule(
            style={"backgroundColor": WBColor.GREY.value, "fontStyle": "italic"},
            condition=[("==", True, "pending")],
        )
        swing_formatting_positive = dp.FormattingRule(
            style={"backgroundColor": WBColor.YELLOW_LIGHT.value},
            condition=[("==", "pos", "swing_pricing_indicator")],
        )
        swing_formatting_negative = dp.FormattingRule(
            style={"backgroundColor": WBColor.BLUE_LIGHT.value},
            condition=[("==", "neg", "swing_pricing_indicator")],
        )
        fields = []

        if "portfolio_id" not in self.view.kwargs:
            fields.append(dp.Field(key="portfolio", label="Portfolio", pinned="left"))

        fields.extend(
            [
                dp.Field(key="value_date", label="Value Date", width=100, pinned="left"),
                dp.Field(
                    key="cash_flow_forecast", label="CF Forecast", width=100, formatting_rules=[negative_formatting]
                ),
                dp.Field(
                    key="total_assets",
                    label="Total AuM",
                    width=100,
                    formatting_rules=[negative_formatting, pending_formatting],
                ),
                dp.Field(
                    key="estimated_total_assets",
                    label="Estimated AuM",
                    width=100,
                    formatting_rules=[negative_formatting],
                ),
                dp.Field(
                    key="cash",
                    label="Cash & E.",
                    width=100,
                    formatting_rules=[negative_formatting, pending_formatting],
                ),
                dp.Field(
                    key="cash_flow_asset_ratio",
                    label="Cash Flow Forecast / E. AuM",
                    width=125,
                    formatting_rules=[negative_formatting, swing_formatting_positive, swing_formatting_negative],
                ),
                dp.Field(key="true_cash", label="True Cash", width=100, formatting_rules=[negative_formatting]),
                dp.Field(key="cash_pct", label="Cash & E. %", width=100, formatting_rules=[negative_formatting]),
                dp.Field(key="true_cash_pct", label="True Cash %", width=100, formatting_rules=[negative_formatting]),
                dp.Field(
                    key="target_cash_pct", label="Target Cash %", width=100, formatting_rules=[negative_formatting]
                ),
                dp.Field(key="target_cash", label="Target Cash", width=100, formatting_rules=[negative_formatting]),
                dp.Field(key="excess_cash", label="Δ Target Cash", width=100, formatting_rules=[negative_formatting]),
                dp.Field(
                    key="proposed_rebalancing",
                    label="Proposed Rbl.",
                    width=100,
                    formatting_rules=[negative_formatting, positive_formatting, bold_formatting],
                ),
                dp.Field(
                    key="rebalancing",
                    label="Rbl.",
                    width=100,
                    formatting_rules=[negative_formatting, positive_formatting, bold_formatting],
                ),
                dp.Field(key="comment", label="Comment", width=300),
            ]
        )
        return dp.ListDisplay(
            fields=fields,
            legends=[
                dp.Legend(items=[dp.LegendItem(label="Pending data", icon=WBColor.GREY.value)]),
                dp.Legend(
                    items=[
                        dp.LegendItem(label="Swing Pricing: Subscription", icon=WBColor.YELLOW_LIGHT.value),
                        dp.LegendItem(label="Swing Pricing: Redemption", icon=WBColor.BLUE_LIGHT.value),
                    ]
                ),
            ],
        )



from datetime import datetime
from io import BytesIO
from typing import Optional

from django.db import models
from pandas.tseries.offsets import BDay
from wbcore.contrib.io.backends import AbstractDataBackend, register
from wbfdm.import_export.backends.refinitiv.utils import Controller
from wbportfolio.import_export.backends.utils import (
    get_timedelta_import_instrument_price,
)

from ..wbfdm.mixin import DataBackendMixin

DEFAULT_MAPPING = {"AX": "factor"}


@register("Adjustment", provider_key="refinitiv")
class DataBackend(DataBackendMixin, AbstractDataBackend):
    CHUNK_SIZE = 50

    def __init__(self, import_credential: Optional[models.Model] = None, **kwargs):
        self.controller = Controller(import_credential.username, import_credential.password)

    def get_files(
        self,
        execution_time: datetime,
        obj_external_ids: list[str] = None,
        **kwargs,
    ) -> BytesIO:
        execution_date = execution_time.date()
        start = kwargs.get("start", (execution_date - BDay(get_timedelta_import_instrument_price())).date())
        fields = list(DEFAULT_MAPPING.keys())
        if obj_external_ids:
            df = self.controller.get_data(obj_external_ids, fields, start, execution_date)
            if not df.empty:
                content_file = BytesIO()
                df.to_json(content_file, orient="records")
                file_name = f"adjustment_chunk-{start:%Y-%m-%d}-{execution_date:%Y-%m-%d}_{datetime.timestamp(execution_time)}.json"
                yield file_name, content_file



import pandas as pd
from django.db.models import F, Sum
from rest_framework import filters
from wbcore.contrib.currency.models import CurrencyFXRates
from wbcore.contrib.io.viewsets import ExportPandasAPIViewSet
from wbcore.contrib.pandas import fields as pf
from wbcore.contrib.permission.permissions import InternalUserPermissionMixin
from wbcore.filters import DjangoFilterBackend
from wbcore.serializers import decorator
from wbcore.utils.strings import format_number
from wbcore.viewsets import ModelViewSet

from wbportfolio.filters import FeesAggregatedFilter, FeesFilter, FeesProductFilterSet
from wbportfolio.models import Fees
from wbportfolio.serializers import FeesModelSerializer

from ..configs import (
    FeeEndpointConfig,
    FeesAggregatedProductPandasDisplayConfig,
    FeesAggregatedProductPandasEndpointConfig,
    FeesAggregatedProductTitleConfig,
    FeesButtonConfig,
    FeesDisplayConfig,
    FeesProductDisplayConfig,
    FeesProductEndpointConfig,
    FeesProductTitleConfig,
    FeesTitleConfig,
)
from ..mixins import UserPortfolioRequestPermissionMixin


class FeesModelViewSet(UserPortfolioRequestPermissionMixin, InternalUserPermissionMixin, ModelViewSet):
    filter_backends = (
        DjangoFilterBackend,
        filters.OrderingFilter,
    )
    queryset = Fees.valid_objects.all()
    serializer_class = FeesModelSerializer
    filterset_class = FeesFilter

    display_config_class = FeesDisplayConfig
    title_config_class = FeesTitleConfig
    button_config_class = FeesButtonConfig
    endpoint_config_class = FeeEndpointConfig
    ordering = ["-fee_date", "id"]  # ordering by id because otherwise there is duplicates in the paginated view

    def get_aggregates(self, queryset, paginated_queryset):
        return {
            "total_value_fx_portfolio": {
                "Σ": format_number(queryset.aggregate(s=Sum("total_value_fx_portfolio"))["s"])
            },
            "total_value_gross_fx_portfolio": {
                "Σ": format_number(queryset.aggregate(s=Sum("total_value_gross_fx_portfolio"))["s"])
            },
        }

    def get_queryset(self):
        if self.is_manager:
            return super().get_queryset().select_related("import_source", "product")
        return Fees.objects.none()


class FeesProductModelViewSet(FeesModelViewSet):
    IDENTIFIER = "wbportfolio:product-fees"

    filterset_class = FeesProductFilterSet

    display_config_class = FeesProductDisplayConfig
    title_config_class = FeesProductTitleConfig
    endpoint_config_class = FeesProductEndpointConfig

    def get_queryset(self):
        if self.is_portfolio_manager:
            return super().get_queryset().filter(product=self.product)

        return Fees.objects.none()


class FeesAggregatedProductPandasView(
    UserPortfolioRequestPermissionMixin, InternalUserPermissionMixin, ExportPandasAPIViewSet
):
    IDENTIFIER = "wbportfolio:aggregetedfees"

    filterset_class = FeesAggregatedFilter

    queryset = Fees.valid_objects.all()

    display_config_class = FeesAggregatedProductPandasDisplayConfig
    title_config_class = FeesAggregatedProductTitleConfig
    endpoint_config_class = FeesAggregatedProductPandasEndpointConfig

    pandas_fields = pf.PandasFields(
        fields=(
            pf.PKField(key="id", label="ID"),
            pf.CharField(key="fee_date", label="Instrument"),
            pf.FloatField(
                key="TRANSACTION",
                label="Transactions",
                decorators=(decorator(decorator_type="text", position="left", value="$"),),
            ),
            pf.FloatField(
                key="PERFORMANCE_CRYSTALIZED",
                label="Performance",
                decorators=(decorator(decorator_type="text", position="left", value="$"),),
            ),
            pf.FloatField(
                key="PERFORMANCE",
                label="Performance Crystalized",
                decorators=(decorator(decorator_type="text", position="left", value="$"),),
            ),
            pf.FloatField(
                key="MANAGEMENT",
                label="Management",
                decorators=(decorator(decorator_type="text", position="left", value="$"),),
            ),
            pf.FloatField(
                key="ISSUER",
                label="Issuer",
                decorators=(decorator(decorator_type="text", position="left", value="$"),),
            ),
            pf.FloatField(
                key="OTHER", label="Other", decorators=(decorator(decorator_type="text", position="left", value="$"),)
            ),
            pf.FloatField(
                key="total", label="Total", decorators=(decorator(decorator_type="text", position="left", value="$"),)
            ),
        )
    )
    ordering = ["-fee_date"]
    ordering_fields = [
        "fee_date",
        "TRANSACTION",
        "PERFORMANCE_CRYSTALIZED",
        "PERFORMANCE",
        "MANAGEMENT",
        "ISSUER",
        "OTHER",
        "total",
    ]

    def get_aggregates(self, request, df):
        if df.empty:
            return {}
        return {
            "TRANSACTION": {"Σ": format_number(df["TRANSACTION"].sum())},
            "PERFORMANCE_CRYSTALIZED": {"Σ": format_number(df["PERFORMANCE_CRYSTALIZED"].sum())},
            "PERFORMANCE": {"Σ": format_number(df["PERFORMANCE"].sum())},
            "MANAGEMENT": {"Σ": format_number(df["MANAGEMENT"].sum())},
            "ISSUER": {"Σ": format_number(df["ISSUER"].sum())},
            "OTHER": {"Σ": format_number(df["OTHER"].sum())},
            "total": {"Σ": format_number(df["total"].sum())},
        }

    def get_queryset(self):
        if self.is_portfolio_manager:
            qs = super().get_queryset().filter(product=self.product)
            return qs.annotate(
                currency_fx_rate_usd=CurrencyFXRates.get_fx_rates_subquery(
                    "fee_date", currency="currency", lookup_expr="exact"
                ),
                total_value_fx_usd=F("total_value") * F("currency_fx_rate_usd"),
            )
        return Fees.objects.none()

    def get_dataframe(self, request, queryset, **kwargs):
        df = pd.DataFrame(queryset.values("fee_date", "transaction_subtype", "total_value_fx_usd"))
        if not df.empty:
            df = (
                df.pivot_table(
                    index="fee_date",
                    columns=["transaction_subtype"],
                    values="total_value_fx_usd",
                    aggfunc="sum",
                )
                .rename_axis(None, axis=1)
                .reset_index()
            )
            for type in Fees.Type.values:
                if type not in df:
                    df[type] = 0
            df["total"] = (
                df["TRANSACTION"]
                + df["PERFORMANCE_CRYSTALIZED"]
                + df["PERFORMANCE"]
                + df["MANAGEMENT"]
                + df["ISSUER"]
                + df["OTHER"]
            )
            df["id"] = df["fee_date"]
            df = df.fillna(0)
        return df
